package com.me.asteroids;

import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Polygon;
import com.badlogic.gdx.math.Vector2;
import com.me.common.Random;

public final class AsteroidFactory {

    public static final Random rand = new Random();

    int vertexCount;
    float size;

    float sizeVariation;
    float angleVariation;

    boolean sizeRelativeToLast;

    public AsteroidFactory setVertexCount(int vertexCount) {
        this.vertexCount = vertexCount;
        return this;
    }

    public AsteroidFactory setSize(float size) {
        this.size = size;
        return this;
    }

    public AsteroidFactory setSizeVariation(float sizeVariation) {
        this.sizeVariation = sizeVariation;
        return this;
    }

    public AsteroidFactory sizeRelativeToLast() {
        this.sizeRelativeToLast = true;
        return this;
    }

    public AsteroidFactory sizeRelativeToInitial() {
        this.sizeRelativeToLast = false;
        return this;
    }

    public AsteroidFactory setAngleVariation(float angleVariation) {
        this.angleVariation = angleVariation;
        return this;
    }

    private void validate() {
        if (vertexCount <= 2) {
            throw new IllegalStateException(String.format("Illegal vertexCount: %d. Must be >= 3.", vertexCount));
        }
        if (size <= 0) {
            throw new IllegalStateException(String.format("Illegal vertexCount: %f. Must be > 0.", size));
        }
        if (sizeVariation < 0) {
            throw new IllegalStateException(String.format("Illegal sizeVariation: %f. Must be >= 0.", sizeVariation));
        }
        if (sizeVariation > size) {
            throw new IllegalStateException(String.format("Illegal sizeVariation: %f. Must be <= size.", sizeVariation));
        }
        if (angleVariation < 0) {
            throw new IllegalStateException(String.format("Illegal angleVariation: %f. Must be >= 0.", angleVariation));
        }
        if (angleVariation >  MathUtils.PI2 / vertexCount / 2) {
            throw new IllegalStateException(String.format("Illegal angleVariation: %f. May cause vertexes positions to swap.", angleVariation));
        }
    }

    private Vector2 applyAngleVariation(Vector2 vertex) {
        if (angleVariation > 0) {
            float half = angleVariation * 0.5f;
            vertex.rotateRad(rand.nextFloat(-half, half));
        }
        return vertex;
    }

    private float applySizeVariation(float size) {
        if (sizeVariation > 0) {
            float half = sizeVariation * 0.5f;
            float variation = rand.nextFloat(-half, half);
            if (sizeRelativeToLast) {
                size += variation;
                size = MathUtils.clamp(size, this.size - half, this.size + half);
            } else {
                size = this.size + variation;
            }
        }
        return size;
    }

    public Polygon generate() {
        validate();

        float angleStep = MathUtils.PI2 / vertexCount;

        // Pick a random starting angle
        float startAngle = rand.nextFloat() * MathUtils.PI2;
        Vector2 dir = new Vector2(MathUtils.cos(startAngle), MathUtils.sin(startAngle));

        float lastSize = size;
        float[] vertices = new float[vertexCount * 2];

        for (int i = 0; i < vertexCount; i++) {
            Vector2 vertex = dir.cpy();

            vertex = applyAngleVariation(vertex);
            lastSize = applySizeVariation(lastSize);

            vertex.scl(lastSize);
            vertices[i * 2] = vertex.x;
            vertices[(i * 2) + 1] = vertex.y;

            dir.rotateRad(angleStep);
        }

        return new Polygon(vertices);
    }

}
